#include <stdio.h>
#include "../include/mpi.h"

/*
    This is a simple program to test the communications performance of
    a parallel machine.  The program tcomm does a more exhaustive
    test of the individual links
 */
    
double round_trip();
double exchange_forcetype();
double exchange_async();
double exchange_sync();

int worker();

int MPI_main(argc,argv)
int argc;
char *argv[];
{
int c;
double (* f)();
long reps,proc1,proc2,len,error_flag;
double t;
long first,last,incr, svals[3];

SYusc_init();

if (SYArgHasName( &argc, argv, 1, "-help" )) {
  if (MPI_getid(0) != 0) return 0;
  fprintf( stderr, "%s - test individual communication speeds\n", argv[0] );
  fprintf( stderr, 
	  "[-sync] [-async] [-force] [-size start end stride] [-reps n]\n" );
  return 0;
  }

reps          = 100;
proc1         = 0;
proc2         = MPI_infog( 0, 0, (int *)0 )-1;
error_flag    = 0;
f             = round_trip;
svals[0]      = 0;
svals[1]      = 1024;
svals[2]      = 32;

if (SYArgHasName( &argc, argv, 1, "-force" )) f = exchange_forcetype;
if (SYArgHasName( &argc, argv, 1, "-async" )) f = exchange_async;
if (SYArgHasName( &argc, argv, 1, "-sync"  )) f = exchange_sync;
SYArgGetIntVec( &argc, argv, 1, "-size", 3, svals );
SYArgGetInt(    &argc, argv, 1, "-reps", &reps );

first = svals[0];
last  = svals[1];
incr  = svals[2];

time_function(reps,first,last,incr,proc1,proc2,f);
}

time_function(reps,first,last,incr,proc1,proc2,f)
long reps,first,last,incr,proc1,proc2;
double (* f)();
{
  long len,myproc;
  double mean_time,rate;
  double t;

  myproc   = MPI_getid(0);
  if (myproc == proc1) 
      printf( "\n#p0\tp1\tlen\ttime\t\tave time (us)\trate\n");

  for(len=first;len<=last;len+=incr){
    GSYNC( ALLPROCS );
    t = (* f) (reps,len,proc1,proc2);
    mean_time = t;
    mean_time = mean_time / reps;            /* take average over trials */
    mean_time = mean_time * (1e6);         /* convert to microseconds */
    rate      = (double)(len)/(mean_time*(1e-6));
    if(myproc==proc1) printf("%d\t%d\t%d\t%f\t%f\t%.2f\n",
                              proc1,proc2,len,t,mean_time,rate);
  }
}

/* Synchronous exchange (head-to-head) */
double exchange_sync(reps,len,proc1,proc2)
long reps,len,proc1,proc2;
{
  double elapsed_time;
  long i,msg_id,myproc;
  char *sbuffer,*rbuffer;
  SYusc_time_t t0, t1;

  sbuffer = (char *)MALLOC( len );
  rbuffer = (char *)MALLOC( len );

  myproc = MPI_getid(0);
  elapsed_time = 0;
  if(myproc==proc1){
      MPI_crecv( "blocking", rbuffer, proc2, 0, len );
      SYusc_clock(&t0);
      for(i=0;i<reps;i++) {
	  MPI_csend( "blocking", sbuffer, proc2, 1, len );
	  MPI_crecv( "blocking", rbuffer, proc2, 1, len );
	  }
      SYusc_clock(&t1);
      elapsed_time = SYuscDiff( &t0, &t1 );
      }

  if(myproc==proc2){
      MPI_csend( "blocking", sbuffer, proc1, 1, len );
      for(i=0;i<reps;i++){
	  MPI_csend( "blocking", sbuffer, proc1, 1, len );
	  MPI_crecv( "blocking", rbuffer, proc1, 1, len );
	  }
  }

  FREE( sbuffer );
  FREE( rbuffer );
  return(elapsed_time);
}

/* Asynchronous exchange (head-to-head) */
double exchange_async(reps,len,proc1,proc2)
long reps,len,proc1,proc2;
{
  double         elapsed_time;
  long           i,myproc;
  int            msg_id;
  char           *sbuffer,*rbuffer;
  SYusc_time_t   t0, t1;

  sbuffer = MALLOC( len );
  rbuffer = MALLOC( len );

  myproc = MPI_getid(0);
  elapsed_time = 0;
  if(myproc==proc1){
    MPI_crecv( "blocking", rbuffer, proc2, 0, len );
    SYusc_clock(&t0);
    for(i=0;i<reps;i++){
      msg_id = MPI_crecv( "nonblocking", rbuffer, proc2, 1, len );
      MPI_csend( "blocking", sbuffer, proc2, 1, len );
      MPI_wait( msg_id );
      }
    SYusc_clock(&t1);
    elapsed_time = SYuscDiff( &t0, &t1 );
    }

  if(myproc==proc2){
      MPI_csend( "blocking", sbuffer, proc1, 0, len );
      for(i=0;i<reps;i++){
	  msg_id = MPI_crecv( "nonblocking", rbuffer, proc1, 1, len );
	  MPI_csend( "blocking", sbuffer, proc1, 1, len );
	  MPI_wait( msg_id );
	  }
      }

  FREE( sbuffer );
  FREE( rbuffer );
  return(elapsed_time);
}

/* head-to-head exchange using forcetypes.  This uses null messages to
   let the sender know when the receive is ready */
double exchange_forcetype(reps,len,proc1,proc2)
long reps,len,proc1,proc2;
{
  double         elapsed_time;
  long           i,myproc, d1, *dmy = &d1;
  ASYNCRecvId_t  msg_id;
  char           *sbuffer,*rbuffer;
  SYusc_time_t   t0, t1;

  sbuffer = MALLOC( len );
  rbuffer = MALLOC( len );

  myproc = MPI_getid(0);
  elapsed_time = 0;
  if(myproc==proc1){
    MPI_crecv( "blocking", rbuffer, proc2, 3, len );
    SYusc_clock(&t0);
    for(i=0;i<reps;i++){
	msg_id = MPI_crecvrr( "nonblocking", rbuffer, proc2, 0, len );
	MPI_csend( "blocking", (char *)0, proc2, 2, 0 );
	MPI_crecv( "blocking", dmy, proc2, 2, 0 );
	MPI_csendrr( "blocking", sbuffer, proc2, 0, len );
	MPI_wait( msg_id );
    }
    SYusc_clock(&t1);
    elapsed_time = SYuscDiff( &t0, &t1 );
  }

  if(myproc==proc2){
      MPI_csend( "blocking", sbuffer, proc1, 3, len );
      for(i=0;i<reps;i++){
	  msg_id = MPI_crecvrr( "nonblocking", rbuffer, proc1, 0, len );
	  MPI_csend( "blocking", (char *)0, proc1, 2, 0 );
	  MPI_crecv( "blocking", dmy, proc1, 2, 0 );
	  MPI_csendrr( "blocking", sbuffer, proc1, 0, len );
	  MPI_wait( msg_id );
	  }
  }

  FREE( sbuffer );
  FREE( rbuffer );
  return(elapsed_time);
}

/* This is a round trip (always unidirectional) */
double round_trip(reps,len,proc1,proc2)
long reps,len,proc1,proc2;
{
  double elapsed_time;
  double mean_time;
  long i,pid,myproc;
  char *rbuffer,*sbuffer;
  SYusc_time_t t0, t1;

  myproc  = MPI_getid( 0 );
  sbuffer = MALLOC( len );
  rbuffer = MALLOC( len );

  elapsed_time = 0;
  if(myproc==proc1){
      MPI_crecv( "blocking", rbuffer, proc2, 0, len );
      SYusc_clock(&t0);
      for(i=0;i<reps;i++){
	  MPI_csend( "blocking", sbuffer, proc2, 1, len );
	  MPI_crecv( "blocking", rbuffer, proc2, 1, len );
	  }
      SYusc_clock(&t1);
      elapsed_time = SYuscDiff( &t0, &t1 );
      }

  if(myproc==proc2){
      MPI_csend( "blocking", sbuffer, proc1, 0, len );
      for(i=0;i<reps;i++){
	MPI_crecv( "blocking", rbuffer, proc1, 1, len );
	MPI_csend( "blocking", sbuffer, proc1, 1, len );
    }
  }

  FREE(sbuffer);
  FREE(rbuffer);
  return(elapsed_time);
}
